/* This is a C program which creates a Web FORM on the fly, based on
 * information submitted by the user.
 *
 * Paul D. Boyle
 * Department of Chemistry
 * Box 8204
 * North Carolina State University
 * Raleigh, NC, 27695-8204
 * email:  boyle@laue.chem.ncsu.edu
 */

#define _POSIX_SOURCE

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <ctype.h>
#include <errno.h>
#include <sys/types.h>


#define MAX_ENTRIES 10000

#define TRUE  1
#define FALSE 0

#define DATABASE_FILE         "/home/wdc10/usa.txt"
#define KEYWORD_SITE          "http://ylp.icpet.nrc.ca/WDC10/keywords.html"
#define BUFFER_SIZE           2048
#define FIELD_DELIMITER       "\t"     /* TAB (0x9) for strtok() parsing */
#define YVONS_NEWLINE_TOKEN   '^'      /* Yvon's token for '\n' in TXT fmt */
#define QUOTE                 0x27     /* Hex ASCII code for ' character */
#define DBL_QUOTE             0x22     /* Hex ASCII code for " character */
#define BLANK                 0x20     /* Hex ASCII code for blank character */

#define N_ITEMS      21 /* number of EnumItems (below) */

/* This list of enum constants, tells which data item is being accessed.
 */
    enum  EnumItems { WORLD, ENTRY, STATE, NA_FA, NA_OT, AD_IN, AD_PO, 
                      PHONE, FAX, TELEX, EMAIL, DE_HI, DE_IN, DE_DA, 
                      DE_FI, BIRTH, CU_PO, IN_KW,
                      IN_O, IUCR, 
                      PERM };

/* Here is an orginal post-query data structure */
typedef struct {
    char *name;
    char *val;
} entry;

/* These prototypes are found in util.c from the httpd1.3R
 * distribution.
 */

char *makeword(char *line, char stop);
char *fmakeword(FILE *f, char stop, int *len);
char x2c(char *what);
void unescape_url(char *url);
void plustospace(char *str);

/* Here are my prototypes */

void CreateWebForm( char *ent[] ); 
int LookUpEntry( char *ent[] );   /* Returns -1 on error, 0 otherwise */
void ParseNewlines( char * );
void ReportErrorHTML( const char * );

int main(int argc, char *argv[])
{
    entry entries[MAX_ENTRIES];
    register int x,m=0;
    int cl;

/* PB added these variables */
    size_t n_bytes;
    char *data[N_ITEMS];
    char world_default[] = "domestic";
    char state_default[] = "NONE";
    char empty_field[] = "?";

    printf("Content-type: text/html%c%c",10,10);

    if(strcmp(getenv("REQUEST_METHOD"),"POST")) {
        printf("This script should be referenced with a METHOD of POST.\n");
        printf("If you don't understand this, see this ");
        printf("<A HREF=\"http://www.ncsa.uiuc.edu/SDG/Software"
        "/Mosaic/Docs/fill-out-forms/overview.html\">forms overview</A>.%c",
                10);
        exit(1);
    }
    if(strcmp(getenv("CONTENT_TYPE"),"application/x-www-form-urlencoded")) {
        printf("This script can only be used to decode form results. \n");
        exit(1);
    }
    cl = atoi(getenv("CONTENT_LENGTH"));

    for(x=0;cl && (!feof(stdin));x++) {
        m=x;
        entries[x].val = fmakeword(stdin,'&',&cl);
        plustospace(entries[x].val);
        unescape_url(entries[x].val);
        entries[x].name = makeword(entries[x].val,'=');
    }

/* From here until the end of the program is largely code written 
 * by Paul D. Boyle
 */
    if( NULL == (data[ENTRY] = malloc(n_bytes = strlen(entries[0].val) + 1) )) {
        perror("malloc()" );
        exit( EXIT_FAILURE );
    }
    memset( data[ENTRY], 0x0, n_bytes );
    strcpy( data[ENTRY], entries[0].val );

    if( NULL == (data[NA_FA] = malloc(n_bytes = strlen(entries[1].val) + 1) )) {
        perror("malloc()" );
        exit( EXIT_FAILURE );
    }
    memset( data[NA_FA], 0x0, n_bytes );
    strcpy( data[NA_FA], entries[1].val );

    if( NULL == (data[NA_OT] = malloc(n_bytes = strlen(entries[2].val) + 1) )) {
        perror("malloc()" );
        exit( EXIT_FAILURE );
    }
    memset( data[NA_OT], 0x0, n_bytes );
    strcpy( data[NA_OT], entries[2].val );

    if( NULL == (data[PERM] = malloc(n_bytes = strlen(entries[3].val) + 1) )) {
        perror("malloc()" );
        exit( EXIT_FAILURE );
    }
    memset( data[PERM], 0x0, n_bytes );
    strcpy( data[PERM], entries[3].val );

    data[WORLD] = &world_default[0];
    data[STATE] = &state_default[0];

    for( x = AD_IN; x <= IUCR; x++ ) {
        if( NULL == (data[x] = malloc( n_bytes
                          = 1 + strlen(empty_field)) ) ) {
            perror("malloc()");
            exit( EXIT_FAILURE ); 
        }
        memset( data[x], 0x0, n_bytes );
        strcpy( data[x], empty_field );
    }
    if( 0 == strcmp( "new", data[ENTRY] ) ) {        /* New entry */

        CreateWebForm( data ); 

        free( data[ENTRY] );
        free( data[NA_FA] );
        free( data[NA_OT] );
        free( data[PERM] );

        exit( EXIT_SUCCESS );
    }
    else {                                           /* Update entry */
        if( -1 == LookUpEntry( data )) {
          ReportErrorHTML( "Your entry could not be located,"
                       " please fill out the form below as a <strong>"
                       "NEW</strong> entry.<p><hr>" );
          data[ENTRY] = realloc( data[ENTRY], n_bytes =
                         strlen( "new" ) + 1 );
          memset( data[ENTRY], 0x0, n_bytes );
          strcpy( data[ENTRY], "new" );
        }
        CreateWebForm( data );
        for( x = NA_FA; x < N_ITEMS; x++ )
                                 free( data[x] );
    }
    exit( EXIT_SUCCESS );
}
/* --- Functions --- */
/* CreateWebForm */
void CreateWebForm( char *ent[] )
{  
    time_t now; 
    char date[17];        /* Array for storing date from strftime() */
    memset(date, 0x0, 17);
    now = time( (time_t)NULL);
    strftime( date, 17, "%d %B %Y", localtime( &now ));
    fprintf( stdout, 
"<html>\n"
"<head>\n"
"<title>\n"
"WDC10 Entry Form\n"
"</title>\n"
"</head>\n"
"<body>\n"
"<h1> World Directory of Crystallographers 10th Edition</h1>\n"
"<h2> Submissions for the\n"
" USA\n"
" Section </h2>\n\n"

"This form is to designed to easily allow crystallographers in the\n\n"

"<strong>USA</strong>\n\n"

" to add or update thier entries in the World\n"
"Directory of Crystallographers (WDC).<p>\n"

"Apart from your full name, some means of getting in touch with you, and\n"
"the permission to distribute, no individual data field is compulsory.  We\n"
"do urge you, however, to fill out the various fields as completely as you\n"
"can because incomplete entries stick out, and make the <i>Directory</i>\n"
"less useful to others. <p>\n\n" );

fprintf( stdout,
"If you recently moved from another (non-North American) country to the"
" United States, you will not be able to use the updatable form.\n"
"Enter yourself as &quot;new&quot; entry.\n"
"<a href=%chttp://www.iucr.ac.uk/docs/wdc9.html%c>Check the contents of your\n"
"WDC 9th edition entry</a>, if you want to refresh your memory.\n"
"<p>\n"
"<h3> Important Note: Permission to Distribute</h3>\n",
DBL_QUOTE, DBL_QUOTE );

fprintf( stdout, 
"The data will not be distributed without your explicit permission. If\n"
"you wish to have your data accessible to others, you <strong>must</strong>\n"
"click the <strong>Yes</strong> button at the bottom of this form. &quot;\n"
"No&quot; is the <em>default</em> reply. <p>\n"
"<hr>\n" );

fprintf( stdout, 
"<FORM METHOD=%cPOST%c action=%chttp://laue.chem.ncsu.edu/cgi-bin/wdc-entry%c>\n",
 DBL_QUOTE, DBL_QUOTE, DBL_QUOTE, DBL_QUOTE );

fprintf( stdout, 
"<INPUT TYPE=%chidden%c NAME=%cworld%c VALUE=%c%s%c>\n",
DBL_QUOTE, DBL_QUOTE, DBL_QUOTE, DBL_QUOTE, DBL_QUOTE, ent[WORLD], DBL_QUOTE );

if( 0 == strcmp( "new", ent[ENTRY] ) ) {
fprintf( stdout,
"<strong>Entry Type:</strong><p>\n"
"<ul>\n"
"<li><INPUT TYPE=%cradio%c NAME=%centry_type%c VALUE=%cnew%c CHECKED>New\n"
"<li><INPUT TYPE=%cradio%c NAME=%centry_type%c VALUE=%cupdate%c >Update\n"
"</ul>\n"
"<p>\n",
DBL_QUOTE, DBL_QUOTE, DBL_QUOTE, DBL_QUOTE, DBL_QUOTE, DBL_QUOTE, 
DBL_QUOTE, DBL_QUOTE, DBL_QUOTE, DBL_QUOTE, DBL_QUOTE, DBL_QUOTE );
}
else {
fprintf( stdout,
"<strong>Entry Type:</strong><p>\n"
"<ul>\n"
"<li><INPUT TYPE=%cradio%c NAME=%centry_type%c VALUE=%cnew%c>New\n"
"<li><INPUT TYPE=%cradio%c NAME=%centry_type%c VALUE=%cupdate%c CHECKED>Update\n"
"</ul>\n"
"<p>\n",
DBL_QUOTE, DBL_QUOTE, DBL_QUOTE, DBL_QUOTE, DBL_QUOTE, DBL_QUOTE, 
DBL_QUOTE, DBL_QUOTE, DBL_QUOTE, DBL_QUOTE, DBL_QUOTE, DBL_QUOTE );
}
fprintf( stdout,
"In order for your entry to be properly processed, you need to specify the\n"
"state postal in which you work.  This is a list of U.S. Postal service \n"
"codes, <strong>select the appropriate state:</strong><br>\n"
"<SELECT NAME=%cstate%c SIZE=5>\n",
DBL_QUOTE, DBL_QUOTE );

fprintf( stdout,
"<OPTION> AK\n"
"<OPTION> AL\n"
"<OPTION> AR\n"
"<OPTION> AZ\n"
"<OPTION> CA\n"
"<OPTION> CO\n"
"<OPTION> CT\n"
"<OPTION> DE\n"
"<OPTION> DC\n"
"<OPTION> FL\n"
"<OPTION> GA\n"
"<OPTION> HI\n"
"<OPTION> IA\n"
"<OPTION> ID\n"
"<OPTION> IL\n"
"<OPTION> IN\n"
"<OPTION> KS\n"
"<OPTION> KY\n"
"<OPTION> LA\n"
"<OPTION> MA\n"
"<OPTION> MD\n"
"<OPTION> ME\n"
"<OPTION> MI\n"
"<OPTION> MN\n"
"<OPTION> MO\n"
"<OPTION> MS\n"
"<OPTION> MT\n"
"<OPTION> NC\n"
"<OPTION> ND\n"
"<OPTION> NE\n"
"<OPTION> NH\n"
"<OPTION> NJ\n"
"<OPTION> NM\n"
"<OPTION> NV\n"
"<OPTION> NY\n"
"<OPTION> OH\n"
"<OPTION> OK\n"
"<OPTION> OR\n"
"<OPTION> PA\n"
"<OPTION> PR\n"
"<OPTION> RI\n"
"<OPTION> SC\n"
"<OPTION> SD\n"
"<OPTION> TN\n"
"<OPTION> TX\n"
"<OPTION> UT\n"
"<OPTION> VA\n"
"<OPTION> VI\n"
"<OPTION> VT\n"
"<OPTION> WA\n"
"<OPTION> WI\n"
"<OPTION> WV\n"
"<OPTION> WY\n"
"<OPTION SELECTED> NONE\n"
"</SELECT>\n"
"<p>\n"
"<hr>\n"
"Whether this is a new or updated entry, please fill out this FORM\n"
"as completely as possible. If you need to use non-ASCII characters,\n"
"please use the CIF codes for these characters.  Consult \n"
"<i>Acta. Cryst.</i>, <b>1996,</b>, <i>B52</i>, 271 for the appropriate\n"
"CIF character codes.<p>\n" );

fprintf( stdout,
"<strong>Family Name:</strong>  <i>e.g.</i> Smith<br>\n"
" <INPUT SIZE=35 NAME=%c_name_family%c VALUE=%c%s%c><p>\n",
DBL_QUOTE, DBL_QUOTE, DBL_QUOTE, ent[NA_FA], DBL_QUOTE );

fprintf( stdout,
"<strong>Title and Forenames:</strong>  <i>e.g.</i> Dr. Mary D.<br>\n"
" <INPUT SIZE=35 NAME=%c_name_other%c VALUE=%c%s%c><p>\n",
DBL_QUOTE, DBL_QUOTE, DBL_QUOTE, ent[NA_OT], DBL_QUOTE );

fprintf( stdout,
"<strong>Name of Institution:</strong><br>\n"
"<i>e.g.</i> Department of Chemistry<br>\n"
"Eastern State University <br>\n"
" <TEXTAREA NAME=%c_address_institution%c ROWS=2 COLS=40 >"
"%s"
"</TEXTAREA> <p>\n",
DBL_QUOTE, DBL_QUOTE, ent[AD_IN]);

fprintf( stdout,
"<strong>Mailing Address:</strong><br>\n"
"<i>e.g.</i>Department of Chemistry<br>\n"
"Eastern State University<br>\n"
"Anytown, NJ, 07932<br>\n"
" <TEXTAREA NAME=%c_address_postal%c ROWS=4 COLS=40 >%s</TEXTAREA><p>\n",
DBL_QUOTE, DBL_QUOTE, ent[AD_PO]);

fprintf( stdout,
"<strong>Phone Number:</strong> (Use CIF format, <i>e.g</i> 1(202)5556431)<br>\n"
" <INPUT SIZE=35 NAME=%c_phone%c VALUE=%c%s%c><p> \n",
DBL_QUOTE, DBL_QUOTE, DBL_QUOTE, ent[PHONE], DBL_QUOTE );

fprintf( stdout,
"<strong>FAX:</strong> (CIF formatted, see above)<br>\n"
" <INPUT SIZE=35 NAME=%c_fax%c VALUE=%c%s%c><p>\n",
DBL_QUOTE, DBL_QUOTE, DBL_QUOTE, ent[FAX], DBL_QUOTE );

fprintf( stdout,
"<strong>WWW URL or Telex:</strong><i>e.g.</i> http://www.esu.edu/~msmith/<br>\n"
" <INPUT SIZE=50 NAME=%c_telex%c VALUE=%c%s%c><p>\n",
DBL_QUOTE, DBL_QUOTE, DBL_QUOTE, ent[TELEX], DBL_QUOTE );

fprintf( stdout,
"<strong>Email Address:</strong>  <i>e.g.</i> msmith@chem.esu.edu<br>\n"
" <INPUT SIZE=40 NAME=%c_email%c VALUE=%c%s%c><p>\n",
DBL_QUOTE, DBL_QUOTE, DBL_QUOTE, ent[EMAIL], DBL_QUOTE );

fprintf( stdout,
"<strong>Highest Degree:</strong>  (Ph.D., M.S. B.S., etc)<br>\n"
" <INPUT SIZE=11 NAME=%c_degree_highest%c VALUE=%c%s%c><p>\n",
DBL_QUOTE, DBL_QUOTE, DBL_QUOTE, ent[DE_HI], DBL_QUOTE );

fprintf( stdout,
"<strong>Granting Institution:</strong> (append country if not the USA)<br>\n"
" <INPUT SIZE=35 NAME=%c_degree_institution%c VALUE=%c%s%c> <p>\n",
DBL_QUOTE, DBL_QUOTE, DBL_QUOTE, ent[DE_IN], DBL_QUOTE );

fprintf( stdout,
"<strong>Year Degree Granted:</strong><br>\n"
" <INPUT SIZE=10 NAME=%c_degree_date%c VALUE=%c%s%c><p>\n",
DBL_QUOTE, DBL_QUOTE, DBL_QUOTE, ent[DE_DA], DBL_QUOTE );

fprintf( stdout,
"<strong>Degree granted in field of:</strong><br>\n"
"<INPUT SIZE=20 NAME=%c_degree_field%c VALUE=%c%s%c><p>\n",
DBL_QUOTE, DBL_QUOTE, DBL_QUOTE, ent[DE_FI], DBL_QUOTE );

fprintf( stdout,
"<strong>Year of Birth:</strong><br>\n"
"<INPUT SIZE=5 NAME=%c_birth_date%c VALUE=%c%s%c><p>\n",
DBL_QUOTE, DBL_QUOTE, DBL_QUOTE, ent[BIRTH], DBL_QUOTE );

fprintf( stdout,
"<strong>Current Position:</strong>  <i>e.g.</i> Staff Crystallographer,\n"
" Professor, etc.<br>\n"
"<INPUT SIZE=35 NAME=%c_current_position%c VALUE=%c%s%c><p>\n",
DBL_QUOTE, DBL_QUOTE, DBL_QUOTE, ent[CU_PO], DBL_QUOTE );

fprintf( stdout,
"<hr>\n"
"<strong>Research Interests.</strong>  A description of your research\n"
"interests must be composed using words from the list of approved key\n"
"words.  If you are unfamiliar with this list, please check out this\n"
"list of \n"
"<a href=%c%s%c><strong>KEYWORDS</strong>.\n"
"</a><p>\n"
"<p>\n"
"<h3>Rules for Keyword Expressions</h3>\n"
"<UL>\n"
"<LI> A juxtaposition of <strong>ANY NUMBER</strong>\n"
" of keywords from the list, in the\n"
"singular or the plural form, <strong>WITH NO CONJUNCTIONS</strong>\n"
" (of, in, whatever)\n"
"and <strong>NO PUNCTUATION</strong> is a valid keyword interest. E.g.\n"
"&quot;NMR_spectroscopy_human_enzymes&quot; is a valid keyword expression this\n"
"year. \n"
"<LI> A keyword expression which was valid last time can be assumed \n"
"to be valid this time.\n"
"<LI> You are allowed <strong>UP TO FOUR</strong>\n"
" of these keyword expressions.<P>\n"
"</UL>\n", DBL_QUOTE, KEYWORD_SITE, DBL_QUOTE );

fprintf( stdout,
" <TEXTAREA NAME=%c_interests_key_words%c ROWS=4 COLS=50 >"
"%s"
"</TEXTAREA><p>\n"
"<p>\n",
DBL_QUOTE, DBL_QUOTE, ent[IN_KW] );

fprintf( stdout,
"You may also enter non-keyword interests.  They will be printed but not\n"
"searched electronically.  You can enter anything you want, but it should\n"
"be reasonably short and should preferably have <em>something</em> to do\n"
"with crystallography.<p>\n"
"<strong>Non-Keyword Interests</strong><br>\n"
" <TEXTAREA NAME=%c_interests_other%c ROWS=4 COLS=50 >%s</TEXTAREA><p>\n",
DBL_QUOTE, DBL_QUOTE, ent[IN_O] );

fprintf( stdout,
"<strong>IUCr Responsibilities:</strong><br>\n"
" <TEXTAREA NAME=%c_iucr_responsibilities%c ROWS=1 COLS=40 >"
"%s"
"</TEXTAREA><p>\n",
DBL_QUOTE, DBL_QUOTE, ent[IUCR] );

fprintf( stdout,
"<hr>\n"
"<h2>IMPORTANT:  Permission to Distribute</h2>\n"
"You <strong>must</strong> click the <strong>Yes</strong> button if\n"
"you want this \n"
"information to be accessible to others.<p>\n"
"I give permission to the IUCr to distribute the above information\n"
"in printed and electronic form:<p>\n" );

if( 0 == strcmp( "no", ent[PERM] ) ) {

fprintf( stdout,
"<INPUT TYPE=%cradio%c NAME=%c_permission_to_distribute%c\n"
"       VALUE=%cno%c CHECKED>No<br>\n"
"<INPUT TYPE=%cradio%c NAME=%c_permission_to_distribute%c\n"
"       VALUE=%cyes%c>Yes<br>\n",
DBL_QUOTE, DBL_QUOTE, DBL_QUOTE, DBL_QUOTE, DBL_QUOTE, DBL_QUOTE,
DBL_QUOTE, DBL_QUOTE, DBL_QUOTE, DBL_QUOTE, DBL_QUOTE, DBL_QUOTE );
}
else {
fprintf( stdout,
"<INPUT TYPE=%cradio%c NAME=%c_permission_to_distribute%c\n"
"       VALUE=%cno%c>No<br>\n"
"<INPUT TYPE=%cradio%c NAME=%c_permission_to_distribute%c\n"
"       VALUE=%cyes%c CHECKED>Yes<br>\n", 
DBL_QUOTE, DBL_QUOTE, DBL_QUOTE, DBL_QUOTE, DBL_QUOTE, DBL_QUOTE,
DBL_QUOTE, DBL_QUOTE, DBL_QUOTE, DBL_QUOTE, DBL_QUOTE, DBL_QUOTE );
}

fprintf( stdout,
"<hr>\n"
"You only need to press the &quot;Submit&quot; button once.<p>\n"
"<INPUT TYPE=%csubmit%c VALUE=%cSubmit Entry%c>\n"
"<INPUT TYPE=%creset%c VALUE=%cReset Values%c>\n"
"</FORM>\n"
"<hr>\n"
"Authored by:<br>\n"
"<address>\n"
"Paul D. Boyle<br>\n"
"boyle@laue.chem.ncsu.edu<br>\n"
"North Carolina State University<br>\n"
"Produced on  %s\n"
"</address>\n"
"</BODY>\n"
"</HTML>\n",
DBL_QUOTE, DBL_QUOTE, DBL_QUOTE, DBL_QUOTE,
DBL_QUOTE, DBL_QUOTE, DBL_QUOTE, DBL_QUOTE, date );
    return;
}
/* LookUpEntry
 * This function must do the following things:
 * 1) open database file
 * 2) find the entry
 * 3) parse the entry and reallocate memory for the *ent[] array
 * 4) copy the fields into the proper array elements
 * 5) close the file
 */
int LookUpEntry( char *ent[] )
{
    FILE *database;                    /* File handle for db file */
    int i,                             /* counting integer */
        n_hits,                        /* count if hit is made */
        buffer_overflow,               /* flags buffer[] overflow */
        not_found;                     /* flage for while() loop */
    size_t n_bytes,                    /* tracks chunk's o' memory */
           buf_size;                   /* tracks buffer size */
    char *ptr;                         /* returned by strtok() op's */
    char *tmp_ptr[N_ITEMS];            /* temp array of pointers */
    char *first_name;                  /* for checking first names */
    char buffer[BUFFER_SIZE];          /* fgets() reads in to this */

    if( NULL == (database = fopen(DATABASE_FILE, "r")) ) {
        perror( DATABASE_FILE );
        exit( EXIT_FAILURE );
    }

    buffer_overflow = FALSE;
    not_found = TRUE;
    n_hits = 0;

    while( not_found && !feof(database) ) {

        fgets( buffer, BUFFER_SIZE, database );
        if( (buf_size = strlen(buffer)) == BUFFER_SIZE )
                                        buffer_overflow = TRUE;

        ptr = strtok( buffer, FIELD_DELIMITER );

        if( 0 == strncmp( ent[NA_FA], ptr, strlen(ent[NA_FA]) ) ) {          /* Check family name */

            n_bytes = 1 + strlen( ptr );
            if( NULL == ( tmp_ptr[NA_FA] = malloc(n_bytes)) ) {
                 perror("malloc()");
                 exit(EXIT_FAILURE);
            }
            memset( tmp_ptr[NA_FA], 0x0, n_bytes );
            strcpy( tmp_ptr[NA_FA], ptr ); 

            ptr = strtok( NULL, FIELD_DELIMITER );

            if( strlen(ent[NA_OT]) <= strlen(ptr) )      /* Check first name */
                first_name = strstr( ptr, ent[NA_OT] );
            else
                first_name = strstr( ent[NA_OT], ptr );

            if( (NULL != first_name) && 
                (FALSE == buffer_overflow) ) {                  /* We found entry */
                n_hits = 1;
                n_bytes = 1 + strlen( ptr );
                if( NULL == ( tmp_ptr[NA_OT] = malloc(n_bytes)) ) {
                     perror("malloc()");
                     exit(EXIT_FAILURE);
                }
                memset( tmp_ptr[NA_OT], 0x0, n_bytes );
                strcpy( tmp_ptr[NA_OT], ptr ); 

                for( i = AD_IN; i < PERM; i++ ) {
                     ptr = strtok( NULL, FIELD_DELIMITER );
                     n_bytes = 1 + strlen( ptr );
                     if( NULL == ( tmp_ptr[i] = malloc(n_bytes)) ) {
                        perror("malloc()");
                         exit(EXIT_FAILURE);
                     }
                     memset( tmp_ptr[i], 0x0, n_bytes );
                     strcpy( tmp_ptr[i], ptr ); 
                }

         /* For the CIF "text" fields insert the '\n' character */
                ParseNewlines( tmp_ptr[AD_IN] );
                ParseNewlines( tmp_ptr[AD_PO] );
                ParseNewlines( tmp_ptr[IN_KW] );
                ParseNewlines( tmp_ptr[IN_O] );
                ParseNewlines( tmp_ptr[IUCR] );

         /* Now copy the entry into the *ent[] array. making proper
          * memory allocations.
          */
                for( i = NA_FA; i < PERM; i++ ) {
                   n_bytes = 1 + strlen( tmp_ptr[i] );
                   if( NULL == ( ent[i] = realloc( ent[i], n_bytes )) ) {
                       perror( "realloc()" );
                       exit( EXIT_FAILURE );
                   }
                   memset( ent[i], 0x0, n_bytes );
                   strcpy( ent[i], tmp_ptr[i] );
                }
                not_found = 0;
                fclose( database ); 
                for( i = NA_FA; i < PERM; i++ )
                                 free( tmp_ptr[i] );
   
            }                         /* close entry found block */
        }                             /* close if( family_name found) block */
    }                                 /* Close while() loop */
    return( n_hits - 1 );
}
/* ParseNewlines */
/* Searches for the newline token of the TXT format and replaces
 * with bona fide '\n' characters.
 */
void ParseNewlines( char *a )
{
    while( NULL != ( a = strchr( a, YVONS_NEWLINE_TOKEN ) ) ) {
         *a = '\n';
         a++;
    } 
    return;   
}
/* ReportErrorHTML */
/* Writes an error message to stdout so the users can see it. */
void ReportErrorHTML( const char *a )
{
    fprintf( stdout, 
    "<html><head><title>\n"
    "Error Report\n"
    "</title></head><body>\n"
    "<h3> Error Report:</h3>\n"
    "%s<p>\n"
    "</body></html>\n", a );
     fflush( stdout );
}


/* the following functions come from Rob McCool's(?) util.c file
 * which came with the httpd-1.3R server source.  I have made some
 * modifications.
*/

#define LF 10
#define CR 13

void getword(char *word, char *line, char stop) {
    int x = 0,y;

    for(x=0;((line[x]) && (line[x] != stop));x++)
        word[x] = line[x];

    word[x] = '\0';
    if(line[x]) ++x;
    y=0;

    while(line[y++] = line[x++]);
}
/* Modify this to do proper error checking and memory initialization
 * in the malloc call -- P. Boyle 22 April 1996 */
char *makeword(char *line, char stop) {
    int x = 0,y;
    size_t n_bytes;
    
/* orig - char *word = (char *) malloc(sizeof(char) * (strlen(line) + 1)); */
    char *word;
    if( NULL ==( word = malloc( n_bytes = sizeof(char) *
                                (strlen(line) + 1 ) ) ) ) {
        perror( "malloc" );
        exit( EXIT_FAILURE );
    }
    memset( word, 0x0, n_bytes );

    for(x=0;((line[x]) && (line[x] != stop));x++)
        word[x] = line[x];

    word[x] = '\0';
    if(line[x]) ++x;
    y=0;

    while(line[y++] = line[x++]);
    return word;
}

char *fmakeword(FILE *f, char stop, int *cl) {
    int wsize;
    size_t n_bytes;
    char *word;
    int ll;

    wsize = 102400;
    ll=0;
/* orig - word = (char *) malloc(sizeof(char) * (wsize + 1)); */
    if( NULL ==( word = malloc( n_bytes = sizeof(char) *
                                (wsize + 1 ) ) ) ) {
        perror( "malloc" );
        exit( EXIT_FAILURE );
    }
    memset( word, 0x0, n_bytes );

    while(1) {
        word[ll] = (char)fgetc(f);
        if(ll==wsize) {
            word[ll+1] = '\0';
            wsize+=102400;
            word = (char *)realloc(word,sizeof(char)*(wsize+1));
        }
        --(*cl);
        if((word[ll] == stop) || (feof(f)) || (!(*cl))) {
            if(word[ll] != stop) ll++;
            word[ll] = '\0';
            return word;
        }
        ++ll;
    }
}

char x2c(char *what) {
    register char digit;

    digit = (what[0] >= 'A' ? ((what[0] & 0xdf) - 'A')+10 : (what[0] - '0'));
    digit *= 16;
    digit += (what[1] >= 'A' ? ((what[1] & 0xdf) - 'A')+10 : (what[1] - '0'));
    return(digit);
}

void unescape_url(char *url) {
    register int x,y;

    for(x=0,y=0;url[y];++x,++y) {
        if((url[x] = url[y]) == '%') {
            url[x] = x2c(&url[y+1]);
            y+=2;
        }
    }
    url[x] = '\0';
}

void plustospace(char *str) {
    register int x;

    for(x=0;str[x];x++) if(str[x] == '+') str[x] = ' ';
}

int rind(char *s, char c) {
    register int x;
    for(x=strlen(s) - 1;x != -1; x--)
        if(s[x] == c) return x;
    return -1;
}

int getline(char *s, int n, FILE *f) {
    register int i=0;

    while(1) {
        s[i] = (char)fgetc(f);

        if(s[i] == CR)
            s[i] = fgetc(f);

        if((s[i] == 0x4) || (s[i] == LF) || (i == (n-1))) {
            s[i] = '\0';
            return (feof(f) ? 1 : 0);
        }
        ++i;
    }
}

void send_fd(FILE *f, FILE *fd)
{
    int num_chars=0;
    char c;

    while (1) {
        c = fgetc(f);
        if(feof(f))
            return;
        fputc(c,fd);
    }
}

int ind(char *s, char c) {
    register int x;

    for(x=0;s[x];x++)
        if(s[x] == c) return x;

    return -1;
}

void escape_shell_cmd(char *cmd) {
    register int x,y,l;

    l=strlen(cmd);
    for(x=0;cmd[x];x++) {
        if(ind("&;`'\"|*?~<>^()[]{}$\\",cmd[x]) != -1){
            for(y=l+1;y>x;y--)
                cmd[y] = cmd[y-1];
            l++; /* length has been increased */
            cmd[x] = '\\';
            x++; /* skip the character */
        }
    }
}

