/* This code is based on the post-query.c file which came with the
 * httpd1.3R distribution.  The mailing capabilities, CIF data structure
 * and other CIF specific features were added by:
 *
 * Paul D. Boyle
 * Department of Chemistry
 * Box 8204
 * North Carolina State University
 * Raleigh, NC, 27695-8204
 * email:  boyle@laue.chem.ncsu.edu
 *
 * RCS version control:
 *$Id: wdc-entry.c,v 0.3 1996/05/02 19:40:06 boyle Exp boyle $
 *
 * The /tmp directory on the system needs to be world writeable for this
 * program to work.
 */
#define _POSIX_SOURCE

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <unistd.h>
#include <errno.h>

#define MAX_ENTRIES 10000

#define QUOTE       0x27               /* Hex ASCII code for ' character */
#define DBL_QUOTE   0x22               /* Hex ASCII code for " character */
#define BLANK       0x20               /* Hex ASCII code for blank character */

#define CHAR_FORMAT "%s\n%c%s%c\n"
#define TEXT_FORMAT "%s\n;%s\n;\n"

#define TMP_FILE "/tmp/request.file"

/* PRE-COMPILE CONFIGURATION:  You will need to change these next group
 * of #define's to fit your local situation.
 */

#define MAIL_COMMAND      "/usr/bin/elm -s 'WDC10 Entry Submission'"
#define EMAIL_HOST        "@pyrite.chem.unc.edu"
#define STATUS_URL  "http://laue.chem.ncsu.edu/web/WDC10/state.html"
#define INTL_STATUS_URL   "http://laue.chem.ncsu.edu/web/WDC10/international.html"
#define EDITION           "WDC10"
#define COUNTRY           "USA" /* Substitute your country here */


/* This list of enum constants, tells which data item is being accessed.
 */
    enum  EnumItems { WORLD, ENTRY, STATE, NA_FA, NA_OT, AD_IN, AD_PO, 
                      PHONE, FAX, TELEX, EMAIL, DE_HI, DE_IN, DE_DA, 
                      DE_FI, BIRTH, CU_PO, IN_KW,
                      IN_O, IUCR, 
                      PERM };

typedef struct {
        enum EnumItems index;
        int type;
        char *comment;
        } CIFDataItems;

/* Here is an orginal post-query data structure */
typedef struct {
    char *name;
    char *val;
} entry;

static char *cif_block[] =
{
"loop_",
"    _name_family                      # char",
"    _name_other                       # char",
"    _address_institution              # text",
"    _address_postal                   # text",
"    _phone                            # char",
"    _fax                              # char",
"    _telex                            # char",
"    _email                            # char",
"    _degree_highest                   # char",
"    _degree_institution               # char",
"    _degree_date                      # char",
"    _degree_field                     # char",
"    _birth_date                       # char",
"    _current_position                 # char",
"    _interests_key_words              # text",
"    _interests_other                  # text",
"    _iucr_responsibilities            # text",
"    _permission_to_distribute         # char",
"#                              See Acta. Cryst. (1992) A48, 949-954",
"#",
""  /* Null string to terminate loop */
};
/* These prototypes are found in util.c from the httpd1.3R
 * distribution.
 */

char *makeword(char *line, char stop);
char *fmakeword(FILE *f, char stop, int *len);
char x2c(char *what);
void unescape_url(char *url);
void plustospace(char *str);

/* Here are my prototypes */
int TestStringValue( char * );

int main(int argc, char *argv[])
{
    entry entries[MAX_ENTRIES];
    register int x,m=0;
    int cl;

/* PB added these variables for the mailer and CIF stuff */

    FILE     *temp_file;
    pid_t    child_pid;
    size_t   n_bytes;
    int      status, delimiter;
    char     *no_entry = "?";
    char     *temp_file_name;
    char     *mail_command_line;
    char     *entry_value;
    char     *country, *url;
    CIFDataItems
           cif_data[] = {
              WORLD, 'd', "# dummy entry",
              ENTRY, 'c', "# entry type",
              STATE, 'c', "# originating state",
              NA_FA, 'c', "# family name",
              NA_OT, 'c', "# title and other name(s)",
              AD_IN, 't', "# institution",
              AD_PO, 't', "# postal address",
              PHONE, 'c', "# phone",
              FAX,   'c', "# fax",
              TELEX, 'c', "# telex or WWW URL",
              EMAIL, 'c', "# email",
              DE_HI, 'c', "# highest degree",
              DE_IN, 'c', "# degree institution",
              DE_DA, 'c', "# degree date",
              DE_FI, 'c', "# degree field",
              BIRTH, 'c', "# year of birth",
              CU_PO, 'c', "# current position",
              IN_KW, 't', "# interests (by keyword)",
              IN_O,  't', "# other interests",
              IUCR,  't', "# IUCr responsibilities",
              PERM,  'c', "# permission to distribute" };

    printf("Content-type: text/html%c%c",10,10);

    if(strcmp(getenv("REQUEST_METHOD"),"POST")) {
        printf("This script should be referenced with a METHOD of POST.\n");
        printf("If you don't understand this, see this ");
        printf("<A HREF=\"http://www.ncsa.uiuc.edu/SDG/Software"
        "/Mosaic/Docs/fill-out-forms/overview.html\">forms overview</A>.%c",
                10);
        exit(1);
    }
    if(strcmp(getenv("CONTENT_TYPE"),"application/x-www-form-urlencoded")) {
        printf("This script can only be used to decode form results. \n");
        exit(1);
    }
    cl = atoi(getenv("CONTENT_LENGTH"));

    for(x=0;cl && (!feof(stdin));x++) {
        m=x;
        entries[x].val = fmakeword(stdin,'&',&cl);
        plustospace(entries[x].val);
        unescape_url(entries[x].val);
        entries[x].name = makeword(entries[x].val,'=');
    }
/* From here until the end of the program is largely code written 
 * by Paul D. Boyle
 */

/* First check to see if the state info is filled out.  If not,
 * punt right away and tell the user.
 * The form uses "NONE" as a the default "state" or "country" and will give
 * an error if the user doesn't select another (proper) state.
 */

   if( 0 == strcmp( entries[STATE].val, "NONE" )  ) {
     fprintf(stdout, "<html>\n");
     fprintf(stdout, "<head><title>\n");
     fprintf(stdout, "%s %s Submission Summary\n", EDITION, country );
     fprintf(stdout, "</title></head><body>");
     fprintf( stdout, "<h2>Input Error.</h2>\n" );
     fprintf( stdout, " You did not select a state from the selection box."
                      " Your entry cannot be further processed. <p>"
                      " <strong> Exiting. </strong><p>"
                      " Please try again.\n" );
     fprintf(stdout, "</body></html>\n");
     exit( EXIT_FAILURE );
   }
/* Decide whether entry is a domestic entry, or of an international
 * origin.
 */
    if( 0 == strcmp( entries[WORLD].val, "international" ) ) { 
        country = entries[STATE].val;
        url = INTL_STATUS_URL;
    }
    else {  /* domestic entry */
        country = COUNTRY;
        url = STATUS_URL;
    }

/* Allocate memory and build up temp_file_name */
    if( NULL == (temp_file_name = malloc(n_bytes = strlen(TMP_FILE) + 8 )) ) {
        perror("Memory Allocation");
        exit(-1);
    }
    memset( temp_file_name, 0x0, n_bytes );

/* Build up temporary file name */

    sprintf(temp_file_name, "%s.%d", TMP_FILE, getpid() );

/* Build up command line for mailer */
    if( NULL == (mail_command_line = malloc(n_bytes = strlen(MAIL_COMMAND) + 
                                      strlen(EMAIL_HOST) +
                                      strlen(temp_file_name) + 3)) ) {
        perror("Memory Allocation");
        exit(-1);
    }
    memset( mail_command_line, 0x0, n_bytes );
    sprintf(mail_command_line, "%s %s%s < %s",
                                MAIL_COMMAND,
                                entries[STATE].val,
                                EMAIL_HOST,
                                temp_file_name );


/* Open a file to use for the mailer */
    if( NULL == (temp_file = fopen( temp_file_name, "w" )) ) {
        perror( temp_file_name );
        exit(-1);
    }


/* Write stuff to the request file */

    fprintf( temp_file,"# ** WDC10 Entry Submission **\n" );
    fprintf( temp_file, "# This submission is a(n) %s entry \n",
                          entries[ENTRY].val );

    fprintf( temp_file, "data_%s                 # %s\n",
                         country, country ); 
    for( x = 0; *cif_block[x]; x++ )
         fprintf(temp_file, "%s\n", cif_block[x] );

    fprintf( temp_file,
            "# Remote Host: %s\n", getenv("REMOTE_HOST") );
    fprintf( temp_file,
            "# Remote IP Address: %s\n", getenv("REMOTE_ADDR") );

    for( x = NA_FA; x <= m; x++ ) {            /* start at family name */
       if( 'c' == cif_data[x].type ) {         /* process char data */
         switch( TestStringValue( entries[x].val ) ) {
            case 0:                            /* string is null */
              entry_value = no_entry;
              delimiter = BLANK;
              break;
            case 1:                            /* string has no quote */
              entry_value = entries[x].val;
              delimiter = QUOTE;
              break;
            case 2:                            /* string has quote */
              entry_value = entries[x].val;
              delimiter = DBL_QUOTE;
              break;
         }
         fprintf( temp_file, CHAR_FORMAT, cif_data[x].comment, delimiter, 
                  entry_value, delimiter );
       }
       else {                                  /* process text data */
         if( 0 == strcmp(entries[x].val, "") )  
           entry_value = no_entry;
         else
           entry_value = entries[x].val;

         if( AD_PO == x ) /* Automatically add COUNTRY to the postal addr */
           fprintf( temp_file, "%s\n;%s\n%s\n;\n", cif_data[x].comment, 
                            entry_value, country );
         else
           fprintf( temp_file,TEXT_FORMAT, cif_data[x].comment,
                            entry_value );
       }
    }
    fprintf( temp_file,
           "\n# Server Protocol: %s\n", getenv("SERVER_PROTOCOL") );
    fprintf( temp_file,
            "# Server Port: %s\n", getenv("SERVER_PORT") );

    fclose( temp_file );

/* fork() a child process off to do the mailing.  Parent does
 * some printing to the user's browser then waits for child to
 * finish before unlinking (deleting) the temp_file.
 */
    switch( child_pid = fork() ) {
        case -1:
           perror("fork");
           exit(-1);
           break;
        case 0:  /* child does this */
           execl( "/bin/sh", "sh", "-c",
                  mail_command_line, (char *) 0 );
           _exit( 127 ); /* Use _exit() to defeat buffer flushing */
           break; 
        default: /* parent does this */

           fprintf(stdout, "<html>\n");
           fprintf(stdout, "<head><title>\n");
           fprintf(stdout, "%s %s Submission Summary\n", EDITION, country );
           fprintf(stdout, "</title></head><body>");
           fprintf(stdout,"<H1>%s %s Submission Summary</H1>",
                            EDITION, country);
           fprintf(stdout,"<p> Watch for an <strong>OK</strong> indicator"
              " by your name at the following URL:<p><strong>%s</strong><p>\n"
            "If the <strong>OK</strong> status does not appear in 3 days,"
            " assume that"
            " something is wrong with your submission.  Please send a"
            " regular mail message to the following addess:<p>%s%s<p>\n",
              url,  entries[STATE].val,EMAIL_HOST );
           fprintf(stdout,
                   "The following data was submitted"
                    " as your entry to the %s %s section:<p>%c", EDITION,
                    country, 10);

           fprintf(stdout,"<ul>%c",10);

           for(x = ENTRY; x <= m; x++)
              fprintf(stdout,"<li> <code>%s = %s</code>%c",
               entries[x].name, entries[x].val,10);

           fprintf(stdout,"</ul>%c",10);

           fprintf(stdout, "</body></html>\n");

/* Wait for child to finish before blowing away file in /tmp */
           if( waitpid( child_pid, &status, 0) != child_pid )
                        perror("waitpid() error"); 

           unlink( temp_file_name );  
           free( temp_file_name );
           free( mail_command_line );
           break;
    } 
    exit( EXIT_SUCCESS );
}
/* Functions */
/* TestStringValue -- returns 0 if NULL, 1 if string has no
 * quote ('), and 2 if string does contain a ' character.
 */
int TestStringValue( char *str )
{
    if( 0 == strcmp( str, "" ) )
      return( 0 );
    else if( NULL == strchr( str, QUOTE ) )
             return( 1 );
         else
             return( 2 );
}
/* the following functions come from Rob McCool's(?) util.c file
 * which came with the httpd-1.3R server source.  I have made some
 * modifications.
*/

#define LF 10
#define CR 13

void getword(char *word, char *line, char stop) {
    int x = 0,y;

    for(x=0;((line[x]) && (line[x] != stop));x++)
        word[x] = line[x];

    word[x] = '\0';
    if(line[x]) ++x;
    y=0;

    while(line[y++] = line[x++]);
}
/* Modify this to do proper error checking and memory initialization
 * in the malloc call -- P. Boyle 22 April 1996 */
char *makeword(char *line, char stop) {
    int x = 0,y;
    size_t n_bytes;
    
/* orig - char *word = (char *) malloc(sizeof(char) * (strlen(line) + 1)); */
    char *word;
    if( NULL ==( word = malloc( n_bytes = sizeof(char) *
                                (strlen(line) + 1 ) ) ) ) {
        perror( "malloc" );
        exit( EXIT_FAILURE );
    }
    memset( word, 0x0, n_bytes );

    for(x=0;((line[x]) && (line[x] != stop));x++)
        word[x] = line[x];

    word[x] = '\0';
    if(line[x]) ++x;
    y=0;

    while(line[y++] = line[x++]);
    return word;
}

char *fmakeword(FILE *f, char stop, int *cl) {
    int wsize;
    size_t n_bytes;
    char *word;
    int ll;

    wsize = 102400;
    ll=0;
/* orig - word = (char *) malloc(sizeof(char) * (wsize + 1)); */
    if( NULL ==( word = malloc( n_bytes = sizeof(char) *
                                (wsize + 1 ) ) ) ) {
        perror( "malloc" );
        exit( EXIT_FAILURE );
    }
    memset( word, 0x0, n_bytes );

    while(1) {
        word[ll] = (char)fgetc(f);
        if(ll==wsize) {
            word[ll+1] = '\0';
            wsize+=102400;
            word = (char *)realloc(word,sizeof(char)*(wsize+1));
        }
        --(*cl);
        if((word[ll] == stop) || (feof(f)) || (!(*cl))) {
            if(word[ll] != stop) ll++;
            word[ll] = '\0';
            return word;
        }
        ++ll;
    }
}

char x2c(char *what) {
    register char digit;

    digit = (what[0] >= 'A' ? ((what[0] & 0xdf) - 'A')+10 : (what[0] - '0'));
    digit *= 16;
    digit += (what[1] >= 'A' ? ((what[1] & 0xdf) - 'A')+10 : (what[1] - '0'));
    return(digit);
}

void unescape_url(char *url) {
    register int x,y;

    for(x=0,y=0;url[y];++x,++y) {
        if((url[x] = url[y]) == '%') {
            url[x] = x2c(&url[y+1]);
            y+=2;
        }
    }
    url[x] = '\0';
}

void plustospace(char *str) {
    register int x;

    for(x=0;str[x];x++) if(str[x] == '+') str[x] = ' ';
}

int rind(char *s, char c) {
    register int x;
    for(x=strlen(s) - 1;x != -1; x--)
        if(s[x] == c) return x;
    return -1;
}

int getline(char *s, int n, FILE *f) {
    register int i=0;

    while(1) {
        s[i] = (char)fgetc(f);

        if(s[i] == CR)
            s[i] = fgetc(f);

        if((s[i] == 0x4) || (s[i] == LF) || (i == (n-1))) {
            s[i] = '\0';
            return (feof(f) ? 1 : 0);
        }
        ++i;
    }
}

void send_fd(FILE *f, FILE *fd)
{
    int num_chars=0;
    char c;

    while (1) {
        c = fgetc(f);
        if(feof(f))
            return;
        fputc(c,fd);
    }
}

int ind(char *s, char c) {
    register int x;

    for(x=0;s[x];x++)
        if(s[x] == c) return x;

    return -1;
}

void escape_shell_cmd(char *cmd) {
    register int x,y,l;

    l=strlen(cmd);
    for(x=0;cmd[x];x++) {
        if(ind("&;`'\"|*?~<>^()[]{}$\\",cmd[x]) != -1){
            for(y=l+1;y>x;y--)
                cmd[y] = cmd[y-1];
            l++; /* length has been increased */
            cmd[x] = '\\';
            x++; /* skip the character */
        }
    }
}

